"""psy-reg: Psyplot plugin for visualizing and calculating regression plots

This package contains the plotters for interactive visualization tasks with the
abilities to calculate and visualize regression plots.
This package uses statsmodels and scipy for it's calculations.
"""
from psy_reg.version import __version__

__author__ = "Philipp Sommer (philipp.sommer@unil.ch)"
